__author__ = "Xun Li <lixun910@gmail.com>"

from ..libgeoda import VecVecDouble, VecPair, VecDouble, VecInt, Pair
from ..libgeoda import gda_azp_greedy, gda_azp_sa, gda_azp_tabu
from ..libgeoda import gda_betweensumofsquare, gda_totalsumofsquare, gda_withinsumofsquare, flat_2dclusters

'''
Changes:
1/20/2021 Add azp_greedy, _sa, azp_tabu
'''

def azp_greedy(p, w, data, **kwargs):
    ''' A greedy algorithm to solve the AZP problem 
    
    Note: 
        The automatic zoning procedure (AZP) was initially outlined in Openshaw (1977) as a way to address some of the consequences of the modifiable areal unit problem (MAUP). In essence, it consists of a heuristic to find the best set of combinations of contiguous spatial units into p regions, minimizing the within sum of squares as a criterion of homogeneity. The number of regions needs to be specified beforehand.

    Arguments:
        p (int): The number of spatially constrained clusters
        w (Weight): an instance of Weight class
        data (tuple):  A list of numeric vectors of selected variable
        bound_variable (tuple, optional): A numeric vector of selected bounding variable
        min_bound (float, optional): A minimum value that the sum value of bounding variable int each cluster should be greater than
        inits (int, optional): The number of construction re-runs, which is for ARiSeL "automatic regionalization with initial seed location"
        init_regions (tuple, optional): The initial regions that the local search starts with. Default is empty. means the local search starts with a random process to "grow" clusters
        scale_method (str, optional): One of the scaling methods {'raw', 'standardize', 'demean', 'mad', 'range_standardize', 'range_adjust'} to apply on input data. Default is 'standardize' (Z-score normalization).
        distance_method (str, optional):  The distance method used to compute the distance betwen observation i and j. Defaults to "euclidean". Options are "euclidean" and "manhattan"
        random_seed (int, optional): The seed for random number generator. Defaults to 123456789. It is the same as GeoDa software
        cpu_threads (int, optional): The number of cpu threads used for parallel computation

    Returns:
        list: A list of numeric vectors represents a group of clusters
    '''
    
    min_bound = 0 if 'min_bound' not in kwargs else kwargs['min_bound']
    bound_variable = [] if 'bound_variable' not in kwargs else kwargs['bound_variable']
    inits = 0 if 'inits' not in kwargs else kwargs['inits']
    init_regions = [] if 'init_regions' not in kwargs else kwargs['init_regions']
    scale_method = "standardize" if "scale_method" not in kwargs else kwargs['scale_method']
    distance_method = 'euclidean' if 'distance_method' not in kwargs else kwargs['distance_method'] 
    random_seed = 123456789 if 'random_seed' not in kwargs else kwargs['random_seed']

    if w.num_obs < 1:
        raise ValueError("The weights is not valid.")

    if p < 1 or p > w.num_obs:
        raise ValueError("The number of clusters should be a positive integer.")

    if len(data) < 1:
        raise ValueError("The data from selected variable is empty.")

    in_data = VecVecDouble()

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    for d in data:
        in_data.push_back(d)

    min_bounds = VecPair()
    max_bounds = VecPair()

    if min_bound > 0 and len(bound_variable) == w.num_obs:
        min_bounds.push_back((min_bound, list(bound_variable)))

    in_init_regions = VecInt(list(init_regions))

    cluster_ids = gda_azp_greedy(p, w.gda_w, in_data, scale_method, inits, min_bounds, max_bounds, in_init_regions, distance_method, random_seed)

    between_ss = gda_betweensumofsquare(cluster_ids, in_data)
    total_ss = gda_totalsumofsquare(in_data)
    ratio = between_ss / total_ss
    within_ss = gda_withinsumofsquare(cluster_ids, in_data)

    return {
        "Total sum of squares" : total_ss,
        "Within-cluster sum of squares" : within_ss,
        "Total within-cluster sum of squares" : between_ss,
        "The ratio of between to total sum of squares" : ratio,
        "Clusters" : flat_2dclusters(w.num_obs, cluster_ids),
    }

def azp_sa(p, w, data, cooling_rate=0.85, **kwargs):
    ''' A simulated annealing algorithm to solve the AZP problem 
    
    Note: 
        The automatic zoning procedure (AZP) was initially outlined in Openshaw (1977) as a way to address some of the consequences of the modifiable areal unit problem (MAUP). In essence, it consists of a heuristic to find the best set of combinations of contiguous spatial units into p regions, minimizing the within sum of squares as a criterion of homogeneity. The number of regions needs to be specified beforehand.
        
    Arguments:
        p (int): The number of spatially constrained clusters
        w (Weight): an instance of Weight class
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        cooling_rate (float): The cooling rate of a simulated annealing algorithm. Defaults to 0.85
        sa_maxit (int, optional): The number of iterations of simulated annealing. Defaults to 1
        bound_variable (tuple, optional): A numeric vector of selected bounding variable
        min_bound (float, optional): A minimum value that the sum value of bounding variable int each cluster should be greater than
        inits (int, optional): The number of construction re-runs, which is for ARiSeL "automatic regionalization with initial seed location"
        init_regions (tuple, optional): The initial regions that the local search starts with. Default is empty. means the local search starts with a random process to "grow" clusters
        scale_method (str, optional): One of the scaling methods {'raw', 'standardize', 'demean', 'mad', 'range_standardize', 'range_adjust'} to apply on input data. Default is 'standardize' (Z-score normalization).
        distance_method (str, optional):  The distance method used to compute the distance betwen observation i and j. Defaults to "euclidean". Options are "euclidean" and "manhattan"
        random_seed (int, optional): The seed for random number generator. Defaults to 123456789. It is the same as GeoDa software
        cpu_threads (int, optional): The number of cpu threads used for parallel computation

    Returns:
        list: A list of numeric vectors represents a group of clusters
    '''

    sa_maxit = 1 if 'sa_maxit'  not in kwargs else kwargs['sa_maxit']
    min_bound = 0 if 'min_bound' not in kwargs else kwargs['min_bound']
    bound_variable = [] if 'bound_variable' not in kwargs else kwargs['bound_variable']
    inits = 0 if 'inits' not in kwargs else kwargs['inits']
    init_regions = [] if 'init_regions' not in kwargs else kwargs['init_regions']
    scale_method = "standardize" if "scale_method" not in kwargs else kwargs['scale_method']
    distance_method = 'euclidean' if 'distance_method' not in kwargs else kwargs['distance_method'] 
    random_seed = 123456789 if 'random_seed' not in kwargs else kwargs['random_seed']

    if cooling_rate <=0 or cooling_rate >=1:
        raise ValueError("Cooling rate should be in range (0,1).")

    if w.num_obs < 1:
        raise ValueError("The weights is not valid.")

    if p < 1 or p > w.num_obs:
        raise ValueError("The number of clusters should be a positive integer.")

    if len(data) < 1:
        raise ValueError("The data from selected variable is empty.")

    in_data = VecVecDouble()

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    for d in data:
        in_data.push_back(d)

    min_bounds = VecPair()
    max_bounds = VecPair()

    if min_bound > 0 and len(bound_variable) == w.num_obs:
        min_bounds.push_back((min_bound, list(bound_variable)))

    in_init_regions = VecInt(list(init_regions))

    cluster_ids = gda_azp_sa(p, w.gda_w, in_data, scale_method, inits, cooling_rate, sa_maxit, min_bounds, max_bounds, in_init_regions, distance_method, random_seed)

    between_ss = gda_betweensumofsquare(cluster_ids, in_data)
    total_ss = gda_totalsumofsquare(in_data)
    ratio = between_ss / total_ss
    within_ss = gda_withinsumofsquare(cluster_ids, in_data)

    return {
        "Total sum of squares" : total_ss,
        "Within-cluster sum of squares" : within_ss,
        "Total within-cluster sum of squares" : between_ss,
        "The ratio of between to total sum of squares" : ratio,
        "Clusters" : flat_2dclusters(w.num_obs, cluster_ids),
    }

def azp_tabu(p, w, data, tabu_length, **kwargs):
    ''' A tabu-search algorithm to solve the AZP problem 
    
    Note: 
        The automatic zoning procedure (AZP) was initially outlined in Openshaw (1977) as a way to address some of the consequences of the modifiable areal unit problem (MAUP). In essence, it consists of a heuristic to find the best set of combinations of contiguous spatial units into p regions, minimizing the within sum of squares as a criterion of homogeneity. The number of regions needs to be specified beforehand.
        
    Arguments:
        p (int): The number of spatially constrained clusters
        w (Weight): an instance of Weight class
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        tabu_length (int): The length of a tabu search heuristic of tabu algorithm. Defaults to 10.
        conv_tabu (int, optional): The number of non-improving moves. Defaults to 10.
        bound_variable (tuple, optional): A numeric vector of selected bounding variable
        min_bound (float, optional): A minimum value that the sum value of bounding variable int each cluster should be greater than
        inits (int, optional): The number of construction re-runs, which is for ARiSeL "automatic regionalization with initial seed location"
        init_regions (tuple, optional): The initial regions that the local search starts with. Default is empty. means the local search starts with a random process to "grow" clusters
        scale_method (str, optional): One of the scaling methods {'raw', 'standardize', 'demean', 'mad', 'range_standardize', 'range_adjust'} to apply on input data. Default is 'standardize' (Z-score normalization).
        distance_method (str, optional):  The distance method used to compute the distance betwen observation i and j. Defaults to "euclidean". Options are "euclidean" and "manhattan"
        random_seed (int, optional): The seed for random number generator. Defaults to 123456789. It is the same as GeoDa software
        cpu_threads (int, optional): The number of cpu threads used for parallel computation

    Returns:
        list: A list of numeric vectors represents a group of clusters
    '''
    
    conv_tabu = 10 if 'conv_tabu' not in kwargs else kwargs['conv_tabu']
    min_bound = 0 if 'min_bound' not in kwargs else kwargs['min_bound']
    bound_variable = [] if 'bound_variable' not in kwargs else kwargs['bound_variable']
    inits = 0 if 'inits' not in kwargs else kwargs['inits']
    init_regions = [] if 'init_regions' not in kwargs else kwargs['init_regions']
    scale_method = "standardize" if "scale_method" not in kwargs else kwargs['scale_method']
    distance_method = 'euclidean' if 'distance_method' not in kwargs else kwargs['distance_method'] 
    random_seed = 123456789 if 'random_seed' not in kwargs else kwargs['random_seed']

    if w.num_obs < 1:
        raise ValueError("The weights is not valid.")
    
    if p < 1 or p > w.num_obs:
        raise ValueError("The number of clusters should be a positive integer.")

    if len(data) < 1:
        raise ValueError("The data from selected variable is empty.")

    in_data = VecVecDouble()

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    for d in data:
        in_data.push_back(d)

    min_bounds = VecPair()
    max_bounds = VecPair()

    if min_bound > 0 and len(bound_variable) == w.num_obs:
        min_bounds.push_back((min_bound, list(bound_variable)))

    in_init_regions = VecInt(list(init_regions))

    cluster_ids = gda_azp_tabu(p, w.gda_w, in_data, scale_method, inits, tabu_length, conv_tabu, min_bounds, max_bounds, in_init_regions, distance_method, random_seed)

    between_ss = gda_betweensumofsquare(cluster_ids, in_data)
    total_ss = gda_totalsumofsquare(in_data)
    ratio = between_ss / total_ss
    within_ss = gda_withinsumofsquare(cluster_ids, in_data)

    return {
        "Total sum of squares" : total_ss,
        "Within-cluster sum of squares" : within_ss,
        "Total within-cluster sum of squares" : between_ss,
        "The ratio of between to total sum of squares" : ratio,
        "Clusters" : flat_2dclusters(w.num_obs, cluster_ids),
    }


from ..libgeoda import gda_demean, gda_standardize, gda_standardize_mad

'''
1/19/2021: add demean, standardize, standardize_mad
'''
__author__ = "Xun Li <lixun910@gmail.com>, "

def demean(data):
    """Demean Standardization

    Note:
        The mean for each variable is subtracting from each observation resulting in mean zero.
    
    Args:
        data (list): An input data of multiple variables for median absolute deviation

    Returns:
        list: A data array
    """
    if data is None:
        raise ValueError("The data from selected variable is empty.")
        
    return gda_demean(data)


def standardize(data):
    """Standardization (Z)

    Note:
        Standarize data by transforming data to have zero mean and unit variance.

    Args:
        data (list): An input data of multiple variables for standardization

    Returns:
        list: A data array
    """
    if data is None:
        raise ValueError("The data from selected variable is empty.")
        
    return gda_standardize(data)

def mad(data):
    """Median Absolute Deviation

    Note:
        Compute the median absolute deviation, i.e., the (lo-/hi-) median of the absolute deviations from the median, and (by default) adjust by a factor for asymptotically normal consistency.

    Args:
        data (list): An input data of multiple variables for standardization

    Returns:
        list: A data array
    """
    if data is None:
        raise ValueError("The data from selected variable is empty.")
        
    return gda_standardize_mad(data)


from ..libgeoda import gda_quantilelisa, gda_multiquantilelisa
from ..libgeoda import VecBool, VecInt, VecVecDouble, VecVecBool, VecDouble
from .lisa import lisa

__author__ = "Xun Li <lixun910@gmail.com>, Hang Zhang <zhanghanggis@163.com>"

'''
Changes:
1/21/2021 update quantile_lisa; add multiquantile_lisa
2/11/2021 update local_quantilelisa, local_multiquantilelisa
'''

def local_quantilelisa(w, data, k, q, **kwargs):
    """Quantile LISA Statistics
    The function to apply quantile LISA statistics

    Args:
        w (Weight): A spatial Weights object
        data (tuple/list/pandas.Series): A list of numeric values of selected variable
        k (int): The number of quantiles, range[1, n-1]
        q (int): The index of selected quantile for lisa, range[0, k-1]
        undefs (list, optional): A list of boolean values to indicate which value is undefined or null
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of lisa class represents the results of quantile lisa
    """
    if w == None:
        raise ValueError("Weights is None.")

    if data is None:
        raise ValueError("The data from selected variable is empty.")

    if k == None or k < 1:
        raise ValueError("The value of k needs to be greater than 1 ")

    if q == None or q < 0:
        raise ValueError("The value of q needs to be greater than 0")
    elif q > k:
        raise ValueError("The value of q needs to be smaller than or equal to k")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']
    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    lisa_obj = gda_quantilelisa(w.gda_w, k, q, list(data), list(undefs), significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return lisa(lisa_obj)

def local_multiquantilelisa(w, data, k, q, **kwargs):
    """Multivariate Quantile LISA Statistics
    The function to apply multivariate quantile LISA statistics

    Args:
        w (Weight): A spatial Weights object
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        k (tuple): A tuple of "k" (int) values indicate the number of quantiles for each variable
        q (tuple): A tuple of "q" (int) values indicate which quantile or interval for each variable used in local join count statistics
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of lisa class represents the results of quantile lisa
    """

    undefs = VecVecBool() if 'undefs' not in kwargs else kwargs['undefs']
    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    if w == None:
        raise ValueError("Weights is None.")

    if len(data) == 0:
        raise ValueError("The input data can not be empty.")

    if len(data) != len(k) or len(k) != len(q):
        raise ValueError("The size of k, q and data are not matched.")

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    lisa_obj = gda_multiquantilelisa(w.gda_w, k, q, data, undefs, significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return lisa(lisa_obj)


from .moran import local_moran, batch_local_moran
from .geary import local_geary, local_multigeary
from .joincount import local_joincount, local_multijoincount, local_bijoincount
from .getisord import local_g, local_gstar
from .lisa import lisa, batchlisa
from .quantile_lisa import local_quantilelisa, local_multiquantilelisa
from .neighbor_match_test import neighbor_match_test


import sys
import string
import random
import math

from .libgeoda import GeoDa, GeoDaTable, VecString, VecBool, VecInt64, VecDouble

import os
__author__ = "Xun Li <lixun910@gmail.com>"
__all__ = ['geoda', 'open']

class geoda:
    """
    A wrapper class of GeoDa class from libgeoda created from ESRI Shapefile

    Attributes:
        num_obs (int): The number of observations
        num_cols (int): The number of columns
        field_names (tuple): A list of field names 
        field_types (dict): A dict of field types 
        map_type (str): The map type (Point, Polygon, LineString)
    """
    def __init__(self, gda_obj):
        """
        Constructor of geoda object.

        Parameters
        ----------
        gda_obj : Object
            An object / pointer of GeoDa class 
        """
        self.gda = gda_obj

        self.num_obs = gda_obj.GetNumObs()

        self.num_cols = gda_obj.GetNumCols()

        self.field_names = self.GetFieldNames()

        self.field_types = self.GetFieldTypes()

        self.map_type = self.GetMapType()

    def GetNumCols(self):
        """
        Get the number of columns

        Returns
        -------
        : int
            the number of columns
        """
        return self.gda.GetNumCols()

    def GetNumObs(self):
        """Get the number of observations

        Return:
            int: thu number of observations
        """
        return self.gda.GetNumObs()

    def GetFieldNames(self):
        """Get the field names of all columns

        Return:
            :obj:`list` of :obj:`str`: a list of field names 
        """
        return self.gda.GetFieldNames()

    def GetFieldTypes(self):
        """Get the field types (integer, real, string) of all columns

        Return:
            :obj:`list` of :obj:`str`: a list of field types
        """
        fnames = self.GetFieldNames()
        ftypes = self.gda.GetFieldTypes()

        newtypes = {}
        for i, ft in enumerate(ftypes):
            fn = fnames[i]
            newtypes[fn] = ft

        return newtypes 

    def GetMapType(self):
        """Get the map type

        Return:
            :obj:`str`: map type
        """
        return self.gda.GetMapTypeName()

    def __getitem__(self, col_name):
        """Get the values from a column using [] operator
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of :obj:`str`: a list of string values of selected column
        """
        if type(col_name) == list:
            result = list()
            ftypes = self.GetFieldTypes()
            for item in col_name:
                ft = ftypes[item]
                if ft == "integer":
                    result.append(self.GetIntegerCol(item))
                elif ft == "real":
                    result.append(self.GetRealCol(item))
                else:
                    result.append(self.GetStringCol(item))
            return result
        else:
            ftypes = self.GetFieldTypes()
            ft = ftypes[col_name]
            if ft == "integer":
                return self.GetIntegerCol(col_name)
            elif ft == "real":
                return self.GetRealCol(col_name)
            else:
                return self.GetStringCol(col_name)

    def GetIntegerCol(self, col_name):
        """Get the integer values from a column
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of int: a list of integer values of selected column
        """
        if not isinstance(col_name, str) or len(col_name) <= 0:
            raise ValueError("The column name is not valid or not existed.")
        return self.gda.GetIntegerCol(col_name)

    def GetRealCol(self, col_name):
        """Get the real values from a column
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of float: a list of float values of selected column
        """
        if not isinstance(col_name, str) or len(col_name) <= 0:
            raise ValueError("The column name is not valid or not existed.")
        return self.gda.GetNumericCol(col_name)

    def GetStringCol(self, col_name):
        """Get the string values from a column
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of :obj:`str`: a list of string values of selected column
        """
        if not isinstance(col_name, str) or len(col_name) <= 0:
            raise ValueError("The column name is not valid or not existed.")
        return self.gda.GetStringCol(col_name)

    def GetUndefinedVals(self, col_name):
        """Get the undefined flags from a column
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of :obj:`bool`: a list of bool flags indicating if the values are undefined of selected column
        """
        if not isinstance(col_name, str) or len(col_name) <= 0:
            raise ValueError("The column name is not valid or not existed.")
        return self.gda.GetUndefinesCol(col_name)

    def __repr__(self):
        info = ""
        info += "geoda object:\n"
        info += "\t Number of observations: {0}\n".format(self.num_obs)
        info += "\t Number of fields: {0}\n".format(self.num_cols)
        info += "\t Geometry type(s): {0}\n".format(self.map_type)
        info += '{0:>24} {1:>28}\n'.format("field name:", "field type (shapfile):") 
        ftypes = self.GetFieldTypes()
        for fn, ft in ftypes.items():
            info += '{0:>24} {1:>28}\n'.format(fn, ft)
        return info

def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))

def geopandas_to_geoda(gdf, with_table=False):
    """Create a geoda instance from geopandas object.

    Note: the table data are NOT copied to pygeoda for performance issue. It is
    recommended to use table data in dataframe directly.

    Args:
        gdf (GeoDataFrame): An instance of geopands class.
        with_table (boolean): A boolean flag indicates if copy the table content to geoda instance
    
    Returns:
        (geoda): An instance of geoda class.
    """

    try:
        import geopandas
    except ImportError:
        print("(Optional) GeoPandas is not found. Please install GeoPandas for ESDA features.")

    geoms = gdf.geometry
    n_rows = len(gdf)
    n_cols = gdf.columns.size
    col_nms = gdf.columns

    # Table
    gda_tbl = GeoDaTable()

    if with_table:
        for i in range(n_cols):
            col_nm = str(col_nms[i])
            if col_nm == 'geometry':
                continue
            col_type = gdf[col_nm].dtype
            vals = gdf[col_nm].to_list()
            if col_type == 'float64' or col_type == 'float':
                vf = VecDouble(n_rows)
                undef = VecBool(n_rows)
                for i in range(n_rows):
                    vf[i] = vals[i]
                    undef[i] = vals[i] == None
                gda_tbl.AddRealColumn(col_nm, vf, undef)
            elif col_type == 'int64' or col_type == 'int':
                vi = VecInt64(n_rows)
                undef = VecBool(n_rows)
                for i in range(n_rows):
                    vi[i] = vals[i]
                    undef[i] = vals[i] == None
                gda_tbl.AddIntColumn(col_nm, vi, undef)
            else:
                vs = VecString(n_rows)
                undef = VecBool(n_rows)
                for i in range(n_rows):
                    undef[i] = vals[i] == None
                    if undef[i] == False:
                        vs[i] = vals[i]
                gda_tbl.AddStringColumn(col_nm, vs, undef)

    # Geoms
    wkb_size = []
    wkb_array = []
    wkb_bytecount = 0
    for i in range(n_rows):
        wkb = geoms[i].wkb
        wkb_array.append(wkb)
        wkb_size.append(len(wkb))
        wkb_bytecount += wkb_size[-1]

    #wkb_bytes = bytes(wkb_bytecount)
    wkb_bytes = bytearray(wkb_bytecount)
    if sys.version_info[0] < 3: wkb_bytes = bytearray()
    start = 0
    for i in range(n_rows):
        wkb_bytes[start: start + wkb_size[i]] = wkb_array[i]
        start += wkb_size[i]
        

    # map type
    if gdf.geom_type[0].endswith("Polygon"):
        map_type = "map_polygons"
    elif gdf.geom_type[0].endswith("Point"):
        map_type = "map_points" 
    elif gdf.geom_type[0].endswith("Line"):
        map_type = "map_lines" 
    else:
        raise ValueError("Error: pygeoda only supports geometry type of Polygon and Point.")

    # random layer name
    layer_name = id_generator()

    # projection will be NOT handled in libgeoda
    gda = GeoDa(gda_tbl, layer_name, map_type, wkb_bytes, wkb_size)

    return geoda(gda)

def geoda_to_geopandas(geoda_obj):
    """Create a geopandas object from a geoda object.

    Args:
        geoda_obj (geoda): An instance of geoda class.
    
    Returns:
        (GeoDataFrame): An instance of geopandas class.
    """
    gda = geoda_obj.gda

    n_cols = gda.GetNumCols()
    col_nms = gda.GetFieldNames()
    col_tps = gda.GetFieldTypes()

    # pandas DF
    data = {}
    for i in range(n_cols):
        c_nm = col_nms[i]
        c_tp = col_tps[i]
        if c_tp == "integer":
            data[c_nm] = gda.GetIntegerCol(c_nm)
        elif c_tp == "numeric":
            data[c_nm] = gda.GetNumericCol(c_nm)
        else:
            if sys.version_info[0] < 3: 
                vals = gda.GetStringCol(c_nm)
                vals = [str(i) for i in vals]
                data[c_nm] = vals
            else:
                data[c_nm] = gda.GetStringCol(c_nm)
    df = pandas.DataFrame(data)

    # geometries
    geoms = []
    wkb_array = gda.GetGeometryWKB()
    for wkb in wkb_array:
        if sys.version_info[0] < 3: wkb = bytearray(wkb)
        shapely_obj = shapely.wkb.loads(bytes(wkb))
        geoms.append(shapely_obj)

    gdf = geopandas.GeoDataFrame(df, geometry=geoms)

    # projection
    return gdf

class geodaGpd(geoda):
    """
    A wrapper class of GeoDa class from libgeoda created from a geopandas object 

    Attributes:
        num_obs (int): The number of observations
        num_cols (int): The number of columns
        field_names (tuple): A list of field names 
        field_types (dict): A dict of field types 
        map_type (str): The map type (Point, Polygon, LineString)
    """
    def __init__(self, gpd_obj):
        self.gp = geopandas_to_geoda(gpd_obj)
        self.gda = self.gp.gda
        self.df = gpd_obj

        self.num_obs = len(self.df)

        self.num_cols = self.GetNumCols()

        self.field_names = self.GetFieldNames()

        self.field_types = self.GetFieldTypes()

        self.map_type = self.GetMapType()

    def GetNumCols(self):
        """
        Get the number of columns

        Returns
        -------
        : int
            the number of columns
        """
        return len(self.df.columns)

    def GetFieldNames(self):
        """Get the field names of all columns

        Return:
            :obj:`list` of :obj:`str`: a list of field names 
        """
        return self.df.columns.tolist()

    def GetFieldTypes(self):
        """Get the field types (integer, real, string) of all columns

        Return:
            :obj:`list` of :obj:`str`: a list of field types
        """
        return self.df.dtypes.to_dict()

    def GetMapType(self):
        """Get the map type

        Return:
            :obj:`str`: map type
        """
        geom_types = tuple(set(self.df.geometry.type))
        return geom_types

    def GetIntegerCol(self, col_name):
        """Get the integer values from a column
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of int: a list of integer values of selected column
        """
        if not isinstance(col_name, str) or len(col_name) <= 0:
            raise ValueError("The column name is not valid or not existed.")

        return self.df[col_name].astype('int64').to_list()

    def GetRealCol(self, col_name):
        """Get the real values from a column
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of float: a list of float values of selected column
        """
        if not isinstance(col_name, str) or len(col_name) <= 0:
            raise ValueError("The column name is not valid or not existed.")

        return self.df[col_name].astype('float64').to_list()

    def GetStringCol(self, col_name):
        """Get the string values from a column
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of :obj:`str`: a list of string values of selected column
        """
        if not isinstance(col_name, str) or len(col_name) <= 0:
            raise ValueError("The column name is not valid or not existed.")

        return self.df[col_name].astype('str').to_list()

    def __getitem__(self, col_name):
        """Get the values from a column using [] operator
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of :obj:`str`: a list of string values of selected column
        """
        if type(col_name) == list:
            return self.df[col_name]
        else:
            return self.df[col_name].to_list()

    def GetUndefinedVals(self, col_name):
        """Get the undefined flags from a column
        Args:
            :obj:`str`: the name of selected column
        Return:
            :obj:`list` of :obj:`bool`: a list of bool flags indicating if the values are undefined of selected column
        """
        if not isinstance(col_name, str) or len(col_name) <= 0:
            raise ValueError("The column name is not valid or not existed.")

        return [i==math.nan for i in self.df[col_name]]

    def __repr__(self):
        info = ""
        info += "geoda object:\n"
        info += "\t Number of observations: {0}\n".format(self.num_obs)
        info += "\t Number of fields: {0}\n".format(self.num_cols)
        info += "\t Geometry type(s): {0}\n".format(self.map_type)
        info += '{0:>24} {1:>30}\n'.format("field name:", "field type (numpy.dtype):") 
        ftypes = self.GetFieldTypes()
        for fn, ft in ftypes.items():
            info += '{0:>24} {1:>30}\n'.format(fn, ft.name)
        return info

def open(data_source):
    """Create a geoda object by reading a spatial dataset: either ESRI Shapefile or GeoPandas object.

    Args:
        data_source (object): The data_source could be either the file path of the ESRI shapefile or a geopandas dataframe object.

    Return:
        :obj:`Object`: An object of geoda instance
    """
    if isinstance(data_source, str):
        ds_path = data_source
        if not isinstance(ds_path, str) or len(ds_path) <= 0:
            raise ValueError("The input path of data source is not valid")
        if not ds_path.lower().endswith('.shp'):
            raise ValueError('Pygeoda can only open ESRI shapefile since v0.0.4')
        if not os.path.exists(ds_path[0:-3]+'dbf'):
            raise ValueError('This shapefile miss a DBF file')
        if not os.path.exists(ds_path[0:-3]+'shx'):
            raise ValueError('This shapefile miss a SHX file')
        gda_obj = GeoDa(ds_path)

        return geoda(gda_obj)
        
    # else try to open a geopandas object
    try:
        import geopandas
        if isinstance(data_source, geopandas.GeoDataFrame):
            return geodaGpd(data_source)
    except:
        raise ValueError("pygeoda can't open current data source. Please use either a file path of an ESRI shapefile or a GeoPandas instance.")



from ..libgeoda import VecVecDouble, VecDouble
from ..libgeoda import gda_redcap
from ..libgeoda import gda_betweensumofsquare, gda_totalsumofsquare, gda_withinsumofsquare, flat_2dclusters

__author__ = "Xun Li <lixun910@gmail.com>, "

def redcap(k, w, data, method, **kwargs):
    ''' Regionalization with dynamically constrained agglomerative clustering and partitioning (REDCAP)

    REDCAP starts from building a spanning tree with 4 different ways (single-linkage, average-linkage, complete-linkage and Ward-linkage).
    Then, REDCAP provides 2 different ways to prune the tree (First-order and Full-order) to build clusters.
    In pygeoda, the following methods are provided:

    * First-order and Single-linkage
    * Full-order and Single-linkage
    * Full-order and Complete-linkage
    * Full-order and Average-linkage
    * Full-order and Ward-linkage

    Arguments:
        k (int): number of clusters
        w (Weight): An instance of Weight class
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        bound_variable (tuple, optional): A numeric vector of selected bounding variable
        min_bound (float, optional): a minimum value that the sum value of bounding variable int each cluster should be greater than 
        scale_method (str, optional): One of the scaling methods {'raw', 'standardize', 'demean', 'mad', 'range_standardize', 'range_adjust'} to apply on input data. Default is 'standardize' (Z-score normalization).
        distance_method (str, optional): {"euclidean", "manhattan"} the distance method used to compute the distance betwen observation i and j. Defaults to "euclidean". Options are "euclidean" and "manhattan"
        random_seed (int,optional): the seed for random number generator. Defaults to 123456789. 
        cpu_threads (int, optional): The number of cpu threads used for parallel computation

    Returns:
        dict: A dict with keys {"Clusters", "TotalSS", "Within-clusterSS", "TotalWithin-clusterSS", "Ratio"}
    '''

    min_bound = 0 if 'min_bound' not in kwargs else kwargs['min_bound']
    bound_variable = [] if 'bound_variable' not in kwargs else kwargs['bound_variable']
    scale_method = "standardize" if "scale_method" not in kwargs else kwargs['scale_method']
    distance_method = 'euclidean' if 'distance_method' not in kwargs else kwargs['distance_method'] 
    random_seed = 123456789 if 'random_seed' not in kwargs else kwargs['random_seed']
    cpu_threads = 6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']

    if method not in ['firstorder-singlelinkage', 'fullorder-singlelinkage', 'fullorder-averagelinkage', 'fullorder-completelinkage', 'fullorder-wardlinkage']:
        raise ValueError('The method has to be one of {"firstorder-singlelinkage", "fullorder-completelinkage", "fullorder-averagelinkage","fullorder-singlelinkage", "fullorder-wardlinkage"}')

    in_data = VecVecDouble()

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    for d in data:
        in_data.push_back(d)
    
    #in_bound_variable = VecDouble(bound_variable)

    cluster_ids = gda_redcap(k, w.gda_w, in_data, scale_method, method, distance_method, bound_variable, min_bound, random_seed, cpu_threads)

    between_ss = gda_betweensumofsquare(cluster_ids, in_data)
    total_ss = gda_totalsumofsquare(in_data)
    ratio = between_ss / total_ss
    within_ss = gda_withinsumofsquare(cluster_ids, in_data)

    return {
        "Total sum of squares" : total_ss,
        "Within-cluster sum of squares" : list(within_ss) + [0]*(len(cluster_ids) - len(within_ss)),
        "Total within-cluster sum of squares" : between_ss,
        "The ratio of between to total sum of squares" : ratio,
        "Clusters" : flat_2dclusters(w.num_obs, cluster_ids),
    }



from ..libgeoda import LISA, BatchLISA
import math

__author__ = "Xun Li <lixun910@gmail.com>, Yeqing Han <yeqinghan_geo@163.com>"

class lisa:
    """A LISA class wrappers the results of LISA computation
    
    Attributes:
        gda_lisa An object of GeoDaLISA
    """

    def __init__(self, lisa_obj):
        """Constructor of lisa object

        Args:
            lisa_obj (Object): An object/pointer of LISA class
        """
        self.gda_lisa = lisa_obj

    def run(self):
        """Call to run LISA computation
        """
        return self.gda_lisa.Run()


    def set_permutations(self, num_perm):
        """Set the number of permutations for the LISA computation

        Args:
            num_perm (int): the number of permutations e.g. 99, 999, 9999, 999999
        """
        if num_perm < 1 or num_perm > 999999:
            raise ValueError("The number of permutations is a positive integer number, but has to be less than 999999.")

        return self.gda_lisa.SetNumPermutations(num_perm)

    def set_threads(self, num_threads):
        """Set the number of CPU threads for the LISA computation

        Args:
            num_threads (int): the number of CPU threads
        """
        if num_threads < 1: 
            raise ValueError("The number of CPU threads has to be a positive integer number.")
            
        return self.gda_lisa.SetNumThreads(num_threads)


    def lisa_values(self):
        """Get LISA values
        Get the local spatial autocorrelation values returned from LISA computation.

        Return:
            tuple: A numeric vector of local spatial autocorrelation
        """
        return self.gda_lisa.GetLISAValues()

    def lisa_pvalues(self):
        """Get pseudo-p values of LISA
        Get the local pseudo-p values of significance returned from LISA computation.

        Return:
            A numeric vector of pseudo-p values of local spatial autocorrelation
        """
        vals = self.gda_lisa.GetLocalSignificanceValues()
        clean_vals = [math.nan if v < 0 else v for v in vals]
        return clean_vals

    def lisa_clusters(self):
        """Get local cluster indicators
        Get the local cluster indicators returned from LISA computation.

        Return:
            A numeric vector of LISA cluster indicator
        """
        return self.gda_lisa.GetClusterIndicators()

    def lisa_num_nbrs(self):
        """Get the number of neighbors of every observations in LISA computation.

        Return:
            :obj:`list` of int: a list of integer values of the number of neighbors of every observations in LISA computation.
                e.g. local moran, local geary etc.
        """
        return self.gda_lisa.GetNumNeighbors()

    def lisa_labels(self):
        """Get the cluster labels of LISA computation.

        Return:
            :obj:`list` of :obj:`str`: a list of string values of the number of the cluster labels of LISA computation
                e.g. local moran, local geary etc.
        """
        return self.gda_lisa.GetLabels()


    def lisa_colors(self):
        """Get the cluster colors of LISA computation.

        Return:
            :obj:`list` of :obj:`str`: a list of string values of the number of the cluster colors of LISA computation
                e.g. local moran, local geary etc.
        """
        return self.gda_lisa.GetColors()

    def lisa_fdr(self, p):
        '''False Discovery Rate value of local spatial autocorrelation
        Get False Discovery Rate value based on current LISA computation and current significant p-value
        
        Args:
            p: A value of current siginificant p-value
            
        Returns:
            float: A numeric vector of False Discovery Rate
        '''
        
        return self.gda_lisa.GetFDR(p)

    def lisa_bo(self, p):
        '''Bonferroni bound value
        Get Bonferroni bound value based on current LISA computation and current significat p-value
        
        Args:
            p: A value of current siginificant p-value
            
        Returns:
            A numeric value of Bonferroni bound
        '''
        
        return self.gda_lisa.GetBO(p)

    def __repr__(self):
        info = "lisa object:\n\n"
        info += "\tlisa_values(): [{0}, ...]\n".format(str(self.lisa_values()[:10])[1:-1])
        info += "\tlisa_pvalues(): [{0}, ...]\n".format(str(self.lisa_pvalues()[:10])[1:-1])
        info += "\tlisa_num_nbrs(): [{0}, ...]\n".format(str(self.lisa_num_nbrs()[:10])[1:-1])
        info += "\tlisa_clusters(): [{0}, ...]\n".format(str(self.lisa_clusters()[:10])[1:-1])
        info += "\tlisa_labels(): {0}\n".format(str(self.lisa_labels()))
        info += "\tlisa_colors(): {0}\n".format(str(self.lisa_colors()))
        return info



class batchlisa:
    """A BatchLISA class wrappers the results of LISA computations
    
    Attributes:
        gda_lisa An object of BatchLISA
    """

    def __init__(self, lisa_obj):
        """Constructor of lisa object

        Args:
            lisa_obj (Object): An object/pointer of LISA class
        """
        self.gda_lisa = lisa_obj

    def run(self):
        """Call to run LISA computation
        """
        return self.gda_lisa.Run()


    def set_permutations(self, num_perm):
        """Set the number of permutations for the LISA computation

        Args:
            num_perm (int): the number of permutations e.g. 999, 9999, 999999
        """
        if num_perm < 1:
            raise ValueError("The number of permutations is a positive integer number.")

        return self.gda_lisa.SetNumPermutations(num_perm)

    def set_threads(self, num_threads):
        """Set the number of CPU threads for the LISA computation

        Args:
            num_threads (int): the number of CPU threads
        """
        if num_threads < 1: 
            raise ValueError("The number of CPU threads has to be a positive integer number.")
            
        return self.gda_lisa.SetNumThreads(num_threads)


    def lisa_values(self, idx):
        """Get LISA values
        Get the local spatial autocorrelation values returned from LISA computation.

        Args:
            idx (int): The index of input data list
        Return:
            tuple: A numeric vector of local spatial autocorrelation
        """
        return self.gda_lisa.GetLISAValues(idx)

    def lisa_pvalues(self, idx):
        """Get the local pseudo-p values of significance returned from LISA computation.

        Args:
            idx (int): The index of input data list

        Return:
            :obj:`list` of float: a list of float values of local pseudo-p values of significance returned from LISA computation
        """
        return self.gda_lisa.GetLocalSignificanceValues(idx)

    def lisa_clusters(self, idx):
        """Get the local cluster indicators returned from LISA computation.

        Args:
            idx (int): The index of input data list

        Return:
            :obj:`list` of float: a list of float values of local cluster indicators returned from LISA computation
        """
        return self.gda_lisa.GetClusterIndicators(idx)

    def lisa_num_nbrs(self):
        """Get numbers of neighbors for all observations
        Get the number of neighbors of every observations in LISA computation.

        Return:
            A numeric vector of the number of neighbors
        """
        return self.gda_lisa.GetNumNeighbors()

    def lisa_labels(self):
        """Get cluster labels
        Get the cluster labels of LISA computation.

        Return:
            A string vector of cluster labels
        """
        return self.gda_lisa.GetLabels()


    def lisa_colors(self):
        """Get cluster colors
        Get the cluster colors of LISA computation.

        Return:
            A string vector of cluster colors
        """
        return self.gda_lisa.GetColors()

    def lisa_fdr(self, p):
        '''False Discovery Rate value of local spatial autocorrelation
        Get False Discovery Rate value based on current LISA computation and current significant p-value
        
        Args:
            p: A value of current siginificant p-value
            
        Returns:
            float: A numeric vector of False Discovery Rate
        '''
        
        return self.gda_lisa.GetFDR(p)

    def lisa_bo(self, p):
        '''Bonferroni bound value
        Get Bonferroni bound value based on current LISA computation and current significat p-value
        
        Args:
            p: A value of current siginificant p-value
            
        Returns:
            float: A numeric value of Bonferroni bound
        '''
        
        return self.gda_lisa.GetBO(p)


from ..libgeoda import VecInt64

__author__ = "Xun Li <lixun910@gmail.com>"

class Weight:
    """
    GeoDa Weight class 
    """
    def __init__(self, gda_w):
        """
        Constructor of Weight class.

        Parameters
        ----------
        gda_w : Object
            A libgeoda GeoDaWeight pointer
        """
        self.gda_w = gda_w
        self.num_obs = gda_w.num_obs

    def is_symmetric(self):
        """
        Check if weights matrix is symmetric

        Parameters
        ----------
        None

        Returns
        -------
        is_symmetric : Boolean
            True means symmetric weights matrix, e.g. contiguity weights
            False means assymmetric weights matrix, e.g. distance-based weights
        """
        if self.gda_w != None:
            return self.gda_w.is_symmetric

    def has_isolates(self):
        """
        Check if any observation has no neighbors 

        Parameters
        ----------
        None

        Returns
        -------
        has_isolates : Boolean
            True means at least one observation has no neighbors 
            False means all observations have at least one neighbor
        """
        if self.gda_w != None:
            return self.gda_w.HasIsolates()

    def weights_sparsity(self):
        """
        Get sparsity computed from weights matrix

        Returns
        -------
        sparsity : float
            sparsity is computed as (#obs w/ neighbor) / (total # obs)
        """
        if self.gda_w != None:
            return self.gda_w.sparsity

    def min_neighbors(self):
        """Get minmum number of neighbors

        Returns:
            int: The number of minimum neighbors 
        """
        if self.gda_w != None:
            return self.gda_w.min_nbrs

    def median_neighbors(self):
        """Get the median number of neighbors

        Returns:
            float: The number of median neighbors 
        """
        if self.gda_w != None:
            return self.gda_w.median_nbrs

    def mean_neighbors(self):
        """Get the mean number of neighbors

        Returns:
            float: The number of mean neighbors 
        """
        if self.gda_w != None:
            return self.gda_w.mean_nbrs

    def max_neighbors(self):
        """Get the maximum number of neighbors

        Returns:
            int: The number of maximum neighbors 
        """
        if self.gda_w != None:
            return self.gda_w.max_nbrs

    def weights_type(self):
        """Get Weights type

        Returns:
            int: weights type
        """
        if self.gda_w != None:
            return self.gda_w.weight_type


    def get_neighbors(self, idx):
        """
        Get neighbors of idx-th observation

        Parameters
        ----------
        idx : int
            The index of observation

        Returns
        -------
        nbrs : tuple
            a tuple of ids of neighbors of specified idx-th observation.
        """
        if self.gda_w != None:
            return self.gda_w.GetNeighbors(idx)

    def spatial_lag(self, values):
        """
        Compute spatial lagged values for values of selected variable 

        Parameters
        ----------
        values : tuple
            The tuple of values of selected column

        Returns
        -------
        lags : list
            the computed spatial lagged values:
        """
        if self.gda_w != None:
            rst = []
            for i in range(self.num_obs):
                rst.append(self.gda_w.SpatialLag(i, values))
            return rst

    def save_weights(self, out_path, layer_name, id_name, id_values):
        """
        Save current spatial weights to a file.

        Parameters
        ----------
        out_path : str
            The path of an output weights file
        layer_name : str
            The name of the layer of input dataset 
        id_name : str
            The id name (or field name), which is an associated column contains unique values, 
            that makes sure that the weights are connected to the correct 
            observations in the data table.
        id_values : tuple
            The tuple of values of selected id_name (column/field)

        """
        if self.gda_w != None:
            return self.gda_w.Save(out_path, layer_name, id_name, id_values)

    def __repr__(self):
        if self.gda_w:
            info = ""
            info += "Weights Meta-data:\n"
            info += '{0:>24} {1:>20}\n'.format("number of observations:", self.num_obs) 
            info += '{0:>24} {1:>20}\n'.format("is symmetric:", "True" if self.is_symmetric() else "False") 
            info += '{0:>24} {1:>20}\n'.format("sparsity:", self.weights_sparsity()) 
            info += '{0:>24} {1:>20}\n'.format("# min neighbors:", self.min_neighbors()) 
            info += '{0:>24} {1:>20}\n'.format("# max neighbors:", self.max_neighbors()) 
            info += '{0:>24} {1:>20}\n'.format("# mean neighbors:", self.mean_neighbors()) 
            info += '{0:>24} {1:>20}\n'.format("# median neighbors:", self.median_neighbors()) 
            info += '{0:>24} {1:>20}\n'.format("has isolates:", "True" if self.has_isolates() else "False") 
            return info